import os
from datetime import datetime, date, timedelta
from peewee import *
from playhouse.sqlite_ext import FTS5Model, SearchField, RowIDField
import sqlite3
import logging

DATABASE_FILENAME = 'library.db'

# Defer init
db = SqliteDatabase(None)


class SyncLog(db.Model):
    timestamp = TimestampField(utc=True)
    status = CharField(default="S")  # S, E, K
    description = TextField(default="")

    def __str__(self):
        return f"[{self.status} {self.timestamp}] {self.description}"


class Series(db.Model):
    tvdb_id = IntegerField(unique=True)
    imdb_id = CharField(null=True)
    name = CharField()
    sort_name = CharField()
    slug = CharField()
    overview = TextField()
    network = CharField(default="")
    poster_url = CharField(default="")
    fanart_url = CharField(default="")
    # ISO code for matched series language in TVDb
    language = CharField(max_length=2, default="en")
    last_updated_on = DateTimeField(default=datetime.utcnow)

    @property
    def imdb_url(self):
        if self.imdb_id:
            return f"https://www.imdb.com/title/{self.imdb_id}/"
        else:
            return ""

    @property
    def tvdb_url(self):
        return f"https://thetvdb.com/series/{self.slug}"

    def __str__(self):
        return self.name


class SeriesIndex(FTS5Model):
    """
    Full-text search index for series
    """
    rowid = RowIDField()
    name = SearchField()
    overview = SearchField()
    network = SearchField()

    class Meta:
        database = db
        options = {'tokenize': 'porter'}


class Tag(db.Model):
    slug = CharField(primary_key=True)
    name = CharField()

    def __str__(self):
        return self.name


class SeriesTag(db.Model):
    series = ForeignKeyField(Series, on_delete="CASCADE")
    # We could change a tag slug, so update this FK accordingly
    tag = ForeignKeyField(Tag, column_name="tag_slug",
                          on_delete="CASCADE", on_update="CASCADE")

    class Meta:
        indexes = (
            (('series', 'tag'), True),
        )


class Episode(db.Model):
    tvdb_id = IntegerField(unique=True)
    series = ForeignKeyField(Series, backref='episodes', on_delete="CASCADE")
    name = CharField()
    season = SmallIntegerField()
    number = SmallIntegerField()
    aired_on = DateField(null=True)
    overview = TextField(default="")
    last_updated_on = DateTimeField(default=datetime.utcnow)
    thumbnail_url = CharField(default="")

    @property
    def season_episode_id(self):
        return "S{:02}E{:02}".format(self.season, self.number)

    def __str__(self):
        return f"{self.season_episode_id} '{self.name}'"

    class Meta:
        indexes = (
            (('series', 'season', 'number'), True),
        )


class EpisodeIndex(FTS5Model):
    """
    Full-text search index for episodes
    """
    rowid = RowIDField()
    name = SearchField()
    overview = SearchField()

    class Meta:
        database = db
        options = {'tokenize': 'porter'}


class Release(db.Model):
    # Enough for BitTorrent 2 SHA-256 hashes
    info_hash = CharField(unique=True, max_length=64)
    # We could change an episode id, so update this FK accordingly
    episode = ForeignKeyField(
        Episode, backref='releases', on_update="CASCADE", on_delete="CASCADE")
    added_on = DateTimeField()
    last_updated_on = DateTimeField()
    size = BigIntegerField()
    magnet_uri = TextField()
    seeders = IntegerField()
    peers = IntegerField()
    name = CharField()
    resolution = SmallIntegerField()

    def __str__(self):
        return self.name


###########
# LOCAL DB QUERIES
###########

def get_last_log():
    """
    Get last successful sync, if any
    """
    return SyncLog.select().where(SyncLog.status == "K").order_by(SyncLog.timestamp.desc()).get_or_none()


def series_subquery():
    SeriesAlias = Series.alias()
    return (SeriesAlias.select(SeriesAlias.id, fn.Max(Episode.season).alias("max_season"))
            .join(Episode)
            .group_by(SeriesAlias.id))


def episode_subquery():
    EpisodeAlias = Episode.alias()
    return (EpisodeAlias.select(EpisodeAlias,  Series.id.alias('series_id'))
            .join(Series)
            )


def get_new_series(interval):
    esubquery = episode_subquery()
    since_date = date.today() - timedelta(days=interval)
    return (Series.select(Series, fn.Count(esubquery.c.id.distinct()).alias("episode_count"))
            .join(Episode)
            .join(Release)
            .join(esubquery, on=(
                esubquery.c.series_id == Series.id))
            .where((Episode.number == 1) &
                   (Episode.season == 1) &
            (Episode.aired_on != None) &
            (Episode.aired_on > since_date) &
            (Release.added_on > since_date))
            .order_by(Release.added_on.desc())
            .group_by(Series.id)
            )


def get_running_series(interval):
    subquery = series_subquery()
    return (Series.select(Series)
            .join(Episode)
            .join(Release)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            .where((subquery.c.max_season == Episode.season) &
            (Release.added_on > (date.today() - timedelta(days=interval))))
            .order_by(Series.sort_name)
            .group_by(Series.id)
            )


def get_featured_series(interval):
    subquery = series_subquery()
    esubquery = episode_subquery()
    return (Series.select(Series, fn.Count(esubquery.c.id.distinct()).alias("episode_count"), fn.SUM(Release.seeders).alias("seeders"))
            .join(Episode)
            .join(Release)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            .join(esubquery, on=(
                esubquery.c.series_id == Series.id))
            .where((subquery.c.max_season == Episode.season)
                   & (Release.added_on > (date.today() - timedelta(days=interval))))
            .order_by(fn.SUM(Release.seeders).desc())
            .group_by(Series.id)
            )


def search_series(query):
    # return SeriesIndex.search(query, weights={'name': 2.0, 'overview': 0.1, 'network': 0.1}, with_score=True, score_alias='search_score')
    return SeriesIndex.search(query, with_score=True, score_alias='search_score')


def get_series(series_id):
    return Series.get(Series.id == series_id)


def get_episode(episode_id):
    return Episode.get(Episode.id == episode_id)


def get_release(id):
    return Release.get(Release.id == id)


# def get_release_with_info_hash(info_hash):
#     return Release.get(Release.info_hash==info_hash)


def get_episodes_for_series(series, season=None):
    subquery = series_subquery()
    return (Episode.select(Episode, fn.Count(Release.info_hash.distinct()).alias('release_count'))
            .join(Release, JOIN.LEFT_OUTER)
            .switch(Episode)
            .join(Series)
            .join(subquery, on=(
                subquery.c.id == Series.id))
            # .where((Episode.series == series.id) & (subquery.c.max_season == Episode.season))
            .where(Episode.series == series.id)
            .group_by(Episode.id)
            .order_by(Episode.season, Episode.number))


def get_releases_for_series(series_id, season=None, max_resolution=2160):
    if season:
        return (Release.select(Release)
                .join(Episode)
                .where((Episode.series == series_id) & (Episode.season == season) & (Release.resolution <= max_resolution))
                .order_by(Episode.season, Episode.number, Release.seeders.desc()))
    else:
        return (Release.select(Release)
                .join(Episode)
                .where((Episode.series == series_id) & (Release.resolution <= max_resolution))
                .order_by(Episode.season, Episode.number, Release.seeders.desc()))


def get_tags_for_series(series):
    return (Tag
            .select()
            .join(SeriesTag)
            .join(Series)
            .where(Series.id == series.id))

# def get_releases_for_episode(episode):
#   return (Release.select()
#       .join(Episode)
#       .where((Episode.id == episode.id))
#       .order_by(Release.seeders))


def connect(app_dir, should_setup=False):
    logging.debug(f"Using SQLite version {sqlite3.sqlite_version}")
    logging.debug(f"Full text search 5? {FTS5Model.fts5_installed()}")

    database = os.path.join(app_dir, DATABASE_FILENAME)
    logging.debug(f"Using local database at {database}")
    db.init(database)
    db.connect()
    if should_setup:
        setup()


def close():
    db.close()


###########
# LOCAL DB SETUP
###########

TAGS = {
    "action": "Action", "adventure": "Adventure", "animation": "Animation", "anime": "Anime", "awards-show": "Awards Show", "children": "Children", "comedy": "Comedy", "crime": "Crime", "documentary": "Documentary", "drama": "Drama", "family": "Family", "fantasy": "Fantasy", "food": "Food", "game-show": "Game Show", "history": "History", "home-and-garden": "Home and Garden", "horror": "Horror", "indie": "Indie", "martial-arts": "Martial Arts", "mini-series": "Mini-Series", "musical": "Musical", "mystery": "Mystery", "news": "News", "podcast": "Podcast", "reality": "Reality", "romance": "Romance", "science-fiction": "Science Fiction", "soap": "Soap", "special-interest": "Special Interest", "sport": "Sport", "suspense": "Suspense", "talk Show": "Talk Show", "thriller": "Thriller", "travel": "Travel", "western": "Western", "war": "War"
}


def setup():
    db.create_tables([
        Series,
        SeriesIndex,
        Episode,
        EpisodeIndex,
        Release,
        Tag,
        SeriesTag,
        SyncLog,
    ], safe=True)

    if Tag.select().count() == 0:
        for slug, name in TAGS.items():
            Tag.create(slug=slug, name=name)



from peewee import chunked
import time
from datetime import datetime, timezone
from requests.exceptions import HTTPError, ReadTimeout
from urllib3.exceptions import ReadTimeoutError
import logging
import videobox.api as api
import videobox.model as model
from videobox.model import SeriesTag, db, Series, SeriesIndex, Episode, EpisodeIndex, Release, SyncLog

INSERT_CHUNK_SIZE = 90      # SQLite has a limit of total 999 max variables
REQUEST_CHUNK_SIZE = 450    # Total URI must be < 4096


class SyncWorker(object):

    def __init__(self, client_id, progress_callback=None, done_callback=None):
        self.client_id = client_id
        self.progress_callback = progress_callback
        self.done_callback = done_callback

    def run(self):
        last_log = model.get_last_log()
        start = time.time()

        # Start log line
        current_log = SyncLog.create(description="Started sync")

        response = []
        if last_log:
            logging.info("Last update done at {0} UTC, requesting updates since then".format(
                last_log.timestamp.isoformat()))
            if self.progress_callback:
                self.progress_callback("Getting updated series...")
            # Ensure UTC tz
            response = self.do_request(lambda: api.get_updated_series(
                self.client_id, last_log.timestamp.replace(tzinfo=timezone.utc)), retries=3)
        else:
            logging.info("No local database found, starting import")
            if self.progress_callback:
                self.progress_callback("First run: import running series...")

            response = self.do_request(
                lambda: api.get_running_series(self.client_id), retries=3)

        if not response:
            # @@TODO pass any error to current_log
            self.update_log(current_log, status="E",
                            description="Unable to contact sync server")
            return

        # Save alert from server, if any
        alert = response["alert"]

        # Grab series
        series_ids = response['series']
        if series_ids:
            logging.debug(
                "Got {0} series, starting update".format(len(series_ids)))
            series_count = self.sync_series(series_ids)

        # Grab episodes
        episode_ids = response['episodes']
        if episode_ids:
            logging.debug(
                "Got {0} episodes, starting update".format(len(episode_ids)))
            episode_count = self.sync_episodes(episode_ids)

        # Grab releases
        release_ids = response['releases']
        if release_ids:
            release_count = self.sync_releases(release_ids)

        elapsed_time = time.time()-start
        if any([series_ids, episode_ids, release_ids]):
            description = "Finished in {:.2f}s: updated {} series, {} episodes, and {} releases".format(
                elapsed_time, series_count, episode_count, release_count
            )
        else:
            description = "Finished in {:.2f}s: no updates were found".format(
                elapsed_time)

        # Mark sync successful
        self.update_log(current_log, status="K", description=description)

        logging.info(f"{description}")

        if self.done_callback:
            self.done_callback(description, alert)

    def sync_series(self, remote_ids):
        instant = datetime.utcnow()

        local_ids = [s.id for s in Series.select(Series.id)]
        new_ids = list(set(remote_ids) - set(local_ids))
        new_ids_count = len(new_ids)

        # Always request all remote ids so we have a change to update existing series
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} series...", 25 + percent)

            logging.debug(
                f"Found missing {new_ids_count} of {len(remote_ids)} series")
            # Reuqest old and new series
            response = self.do_chunked_request(
                api.get_series_with_ids, remote_ids, callback)
            if response:
                with db.atomic():
                    logging.debug("Saving series to database...")
                    for batch in chunked(response, INSERT_CHUNK_SIZE):
                        # Insert new series and attempt to update existing ones
                        (Series.insert_many(batch)
                         .on_conflict(
                            conflict_target=[Series.id],
                            # Pass down values from insert clause
                            preserve=[Series.name, Series.overview, Series.network, Series.poster_url, Series.fanart_url],
                            update={Series.last_updated_on: instant})
                         .execute())
                        for series in batch:
                            (SeriesIndex.insert({
                                SeriesIndex.rowid: series['id'],
                                SeriesIndex.name: series['name']
                            })
                                # Just replace name edits
                                .on_conflict_replace()
                                .execute())
                SeriesIndex.optimize()

            #  Series tags
            response = self.do_chunked_request(
                api.get_series_tags_for_ids, remote_ids, callback)
            if response:
                with db.atomic():
                    logging.debug("Saving series tags to database...")
                    for batch in chunked(response, INSERT_CHUNK_SIZE):
                        (SeriesTag.insert_many(batch)
                            .on_conflict(action="nothing", conflict_target=[SeriesTag.series, SeriesTag.tag])
                         .execute())

        return len(remote_ids)

    def sync_episodes(self, remote_ids):
        instant = datetime.utcnow()

        local_ids = [e.id for e in Episode.select(Episode.id)]
        new_ids = list(set(remote_ids) - set(local_ids))
        new_ids_count = len(new_ids)

        # Always request all remote ids so we have a change to update existing episodes
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} episodes...", 50 + percent)

            logging.debug(
                f"Found missing {new_ids_count} of {len(remote_ids)} episodes")
            # Request old and new episodes
            response = self.do_chunked_request(
                api.get_episodes_with_ids, remote_ids, callback)
            if response:
                with db.atomic():
                    logging.debug("Saving episodes to database...")
                    for batch in chunked(response, INSERT_CHUNK_SIZE):
                        # We need to cope with the unique constraint for (series, season, number)
                        #   index because we cannot rely on episodes id's,
                        #   they are often changed when TVDB users update them
                        (Episode
                            .insert_many(batch)
                            .on_conflict(
                                conflict_target=[
                                    Episode.series, Episode.season, Episode.number],
                                # Pass down values from insert clause 
                                preserve=[Episode.name, Episode.overview, Episode.aired_on, Episode.thumbnail_url],
                                update={Episode.last_updated_on: instant})
                            .execute())
                        # EpisodeIndex.insert({
                        #     EpisodeIndex.rowid: episode_id,
                        #     EpisodeIndex.name: episode.name,
                        #     EpisodeIndex.overview: episode.overview}).execute()

        return len(remote_ids)

    def sync_releases(self, remote_ids):

        local_ids = [r.id for r in Release.select(Release.id)]
        new_ids = list(set(remote_ids) - set(local_ids))
        new_ids_count = len(new_ids)

        # Always request all remote ids so we have a chance to update existing releases
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} releases...", 75 + percent)

            logging.debug(
                f"Found missing {new_ids_count} of {len(remote_ids)} releases")
            # Request old and new releases
            response = self.do_chunked_request(
                api.get_releases_with_ids, remote_ids, callback)
            if response:
                with db.atomic():
                    logging.debug("Saving releases to database...")
                    for batch in chunked(response, INSERT_CHUNK_SIZE):
                        (Release
                            .insert_many(batch)
                            .on_conflict(
                                conflict_target=[Release.id],
                                # Pass down values from insert clause
                                preserve=[Release.peers, Release.seeders, Release.last_updated_on])
                            .execute())

        return len(remote_ids)

    def progress(self, value, min, max):
        return self.scale_between(value, 0, 25, min, max)

    def scale_between(self, value, min_allowed, max_allowed, min, max):
        return (max_allowed - min_allowed) * (value - min) / (max - min) + min_allowed

    def update_log(self, log, status, description=""):
        log.status = status
        log.description = description
        log.save()

    def do_chunked_request(self, handler, ids, callback=None):
        result = []
        ids_count = len(ids)
        for index, chunked_ids in enumerate(chunked(ids, REQUEST_CHUNK_SIZE)):
            if callback:
                percent = self.progress(index*REQUEST_CHUNK_SIZE, 0, ids_count)
                callback(percent, ids_count -
                         index*REQUEST_CHUNK_SIZE)
            logging.debug(
                f"Requesting {index + 1} of {ids_count // REQUEST_CHUNK_SIZE + 1} chunks")
            result.extend(self.do_request(
                lambda: handler(self.client_id, chunked_ids)))
        return result

    def do_request(self, handler, retries=1):
        for index in reversed(range(retries)):
            try:
                response = handler()
                response.raise_for_status()  # Raise an exeption on HTTP errors
            except Exception as ex:
                self.log_network_error(ex, index)
                time.sleep(2)
                continue  # Next retry
            return response.json()
        return []

    def log_network_error(self, ex, retry):
        if isinstance(ex, TimeoutError) or isinstance(ex, ReadTimeoutError) or isinstance(ex, ReadTimeout):
            logging.warn(
                f'Server timed out while handling the request {ex}, {"retrying" if retry else "skipped"}')
        elif isinstance(ex, HTTPError):
            logging.error(
                f'A server error occured while handling the request {ex}, {"retrying" if retry else "skipped"}')
        else:
            # Cannot handle this
            raise ex



import os
import sys
import logging
from pathlib import Path
from datetime import date, datetime, timedelta
from optparse import OptionParser, OptionGroup, Option, make_option
import configparser
import uuid
import subprocess
import itertools
import textwrap
from termcolor import cprint, colored
import videobox
import videobox.sync as sync
import videobox.model as model
import videobox.utilities as utilities


MIN_SEEDERS = 5
SERIES_RUNNING_DAYS = 30

package_dir = Path(__file__).parent


def running_command(args, options):
    results = model.get_running_series(options.days)
    print(
        f"Found {len(results)} series updated in the last {options.days} days:")
    initials_series = itertools.groupby(results, key=get_initial)
    for initial, series in initials_series:
        cprint(f"\n{initial}", attrs=["bold"])
        for s in series:
            print(f" · {s.sort_name}  {colored(s.network,  'light_grey')}")
    # print(f"Use {colored('videobox search name', 'light_grey')} to found out more")


def get_initial(series):
    value = (series.sort_name[0]).upper()
    if value.isalpha():
        return value
    elif value.isdigit():
        return "#"
    else:
        return "?"


def update_command(config):
    worker = sync.SyncWorker(
        config.get("api", "client_id"), progress_callback=on_update_progress, done_callback=on_update_done)
    worker.run()


def on_update_progress(message, percent=0):
    bar_remaining = "." * (int(100 - percent) // 5)
    bar_done = "#" * (int(percent) // 5)
    print(f"{bar_done}{bar_remaining} {message:40}\r", end="")


def on_update_done(message, alert):
    print(f"{message:60}")
    if alert:
        cprint(alert, "yellow")


def download_command(parser, args, options):
    query = sanitize_query(" ".join(args))
    if not query:
        parser.error("missing download query.")

    try:
        max_resolution = int(options.max_resolution.replace("p", "", 1))
    except ValueError:
        parser.error(f"unrecognized resolution {options.max_resolution}")

    download_dir = options.output_dir
    logging.debug(f"Download dir is {download_dir}")
    results = model.search_series(query)
    if len(results) == 1:
        # Single match
        series = results[0]
        releases = model.get_releases_for_series(
            series.rowid, options.season, max_resolution)
        best_releases = find_best_releases(releases)
        if options.dry_run:
            print(
                f"Ready to download {len(best_releases)} releases for {f'season {options.season} of ' if options.season else ''}series '{series.name}' into {download_dir}:\n")
        else:
            print(
                f"Start downloading {len(best_releases)} releases for {f'season {options.season} of ' if options.season else ''}series '{series.name}' into {download_dir} via aria2c...\n")

        print("Seeds  Res.   Size Name")
        cprint("-" * 80, "dark_grey")
        for _, r in best_releases:
            print_release(r)

        if not options.dry_run:
            completed_process = run_aria2(
                download_dir, [r.magnet_uri for _, r in best_releases])
    elif results:
        # Multiple matches
        print(
            f"Found {len(results)} series matching query '{query}':\n")
        for r in results:
            print(
                f" · {r.name}")
        print("\nPlease restrict your query.")
    else:
        print(
            f"No series to download matching '{query}'.")


def run_aria2(download_dir, magnet_uris):
    try:
        with open(package_dir.joinpath("trackers.txt")) as f:
            trackers = f.read()
    except FileNotFoundError as ex:
        logging.warning(
            "Could not open trackers file, using magnet URI's metadata only")
        trackers = ""

    args = [
        "aria2c",
        "--seed-time=0",
        "--allow-overwrite=false",
        f"-d {download_dir}"
    ]
    args.extend(magnet_uris)
    # Move tracker list on tail to avoid messing up things
    args.append(f"--bt-tracker={trackers}")
    # process = subprocess.run(args, capture_output=False, text=True)
    logging.debug(f"Run {' '.join(args)}")
    completed_process = subprocess.run(args)
    return completed_process


def find_best_releases(releases):
    grouped = itertools.groupby(
        releases, lambda r: r.episode.season_episode_id)
    best_releases = []
    for key, episode_releases in grouped:
        release = find_best_release_for_episode(episode_releases)
        best_releases.append((key, release))
    return best_releases


def find_best_release_for_episode(releases):
    # The higher resolution the best
    releases = sorted(releases, key=lambda r: (
        r.resolution, r.size, r.seeders), reverse=True)
    return releases[0]


def print_release(r):
    if r.seeders < MIN_SEEDERS:
        cprint(f"{r.seeders:5} {print_resolution(r.resolution, False)} {utilities.format_size(r.size)} {r.name}", "dark_grey")
    else:
        print(f"{r.seeders:5} {print_resolution(r.resolution)} {utilities.format_size(r.size)} {colored(r.name, 'light_grey')}")


def print_resolution(resolution, use_color=True):
    if use_color:
        colored_ = colored
    else:
        def colored_(x, _): return x  # Uncolored
    if resolution == 2160:
        return colored_(f"{resolution:4}p", "magenta")
    elif resolution == 1080:
        return colored_(f"{resolution:4}p", "yellow")
    elif resolution == 720:
        return colored_(f"{resolution:4}p", "cyan")
    elif resolution == 480:
        return f"{resolution:4}p"
    else:
        return "-"*5


def search_command(parser, args, options):
    query = sanitize_query(" ".join(args))
    if not query:
        parser.error("missing search query.")
    results = model.search_series(query)
    if len(results) == 1:
        today = date.today()
        # Single match
        series = model.get_series(results[0].rowid)
        tags = model.get_tags_for_series(series)
        episodes = model.get_episodes_for_series(series)
        episode_count = len(episodes)
        seasons = set([episode.season for episode in episodes])
        release_count = sum([episode.release_count for episode in episodes])
        cprint("-" * 30, "dark_grey")
        cprint(
            f"{colored(f'{series.name}', attrs=['bold'])}  {colored(series.network, 'light_grey')}")
        cprint("-" * 30, "dark_grey")
        if series.overview:
            print("\n".join(textwrap.wrap(series.overview, 60)))
        print(" ".join([colored(f"#{t.name}", "light_grey") for t in tags]))
        print(f"\nFound {len(seasons)} {plural('season', len(seasons))} with a total of {episode_count} {plural('episode', episode_count)} and {release_count} {plural('release', release_count)}:")
        for season, episodes in itertools.groupby(episodes, key=lambda e: e.season):
            cprint(f"\nSeason {season}", attrs=["bold"])
            for e in episodes:
                print_episode(e, today)
        print(f"\nMore info:")
        print(f" · {series.tvdb_url}")
        if series.imdb_url:
            print(f" · {series.imdb_url}")
    elif results:
        # Multiple matches
        print(
            f"Found {len(results)} series matching query '{query}':\n")
        for search_result in results:
            print(
                f" · {search_result.name}")
    else:
        print(
            f"No series found matching '{query}'. Perhaps try to relax the search query a little?")


def print_episode(e, today):
    aired = ""
    if e.aired_on and e.aired_on < today:
        # Past
        aired = f"Aired {utilities.datetime_since(e.aired_on, today)}"
    elif e.aired_on and e.aired_on >= today:
        # Future
        aired = f"Will air on {utilities.format_date(e.aired_on)}"
    label = f" {e.number:2} {e.name}  {colored(aired, 'light_grey')}{colored(f' with {e.release_count} releases' if e.release_count > 0 else '', 'light_grey')}"
    cprint(label)


def plural(prefix, value):
    return f"{prefix}{'s' if value > 1 else ''}"


def sanitize_query(query):
    # https://www.sqlite.org/fts5.html
    sanitized_query = ""
    for c in query:
        if c.isalpha() or c.isdigit() or ord(c) > 127 or ord(c) == 96 or ord(c) == 26:
            # Allowed in FTS queries
            sanitized_query += c
        else:
            sanitized_query += " "
    return sanitized_query


def auto_update_if_stale(config):
    last_log = model.get_last_log()
    if (not last_log) or (datetime.utcnow() - last_log.timestamp) > timedelta(days=1):
        print("Local database is stale, performing auto-update...")
        update_command(config)


def run_command(config, parser, name, args, options):
    if name == "update":
        update_command(config)
    elif name == "search":
        auto_update_if_stale(config)
        search_command(parser, args, options)
    elif name == "download":
        auto_update_if_stale(config)
        download_command(parser, args, options)
    elif name == "running":
        auto_update_if_stale(config)
        running_command(args, options)
    else:
        raise CommandNotFound()

#############
# Configuration
#############


INI_FILENAME = 'videobox.ini'


def save_config(app_dir, config):
    with open(os.path.join(app_dir, INI_FILENAME), 'w') as out:
        config.write(out)


def load_config(app_dir):
    config = configparser.ConfigParser()
    config.read(os.path.join(app_dir, INI_FILENAME))
    return config


#############
# CLI
#############

class CommandNotFound(RuntimeError):
    pass


COMMANDS = [
    ('download',    'Download all available series releases or a single season'),
    ('running',     'List all series with new releases in the last 30 days'),
    ('search',      'Search for a series by name'),
    ('update',      'Update local database')
]

EPILOG = ""
USAGE = '%prog command [options] [query]\n\n\
Available commands are:\n\n' + "\n".join(f"  {c}\t{help}" for c, help in COMMANDS)


def make_parser():
    parser = OptionParser(prog="videobox",
                          version=videobox.__version__, usage=USAGE, epilog=EPILOG)

    parser.add_option('-v', '--verbose', dest='verbose',
                      help='set logging to DEBUG level', action="store_true", default=False)

    download_options = OptionGroup(parser, "Download options")
    download_options.add_option('-s', '--season',
                                dest='season', type='int', help='download only the given season number')
    download_options.add_option('-d', '--dir',
                                dest='output_dir', help='directory to save downloaded files (default: current working directory)', default=os.getcwd())
    download_options.add_option('--dry-run',
                                dest='dry_run', help='do not download anything, just list what would be', action="store_true")
    download_options.add_option('-r', '--max-resolution',
                                dest='max_resolution', help='limit downloads to 1080p, 720p, or 480p videos (default: 2160p or highest found)', default="2160p")
    parser.add_option_group(download_options)

    search_options = OptionGroup(parser, "Search options")
    search_options.add_option('-y', '--days',
                              dest='days', type='int', help=f'show series updated since number of days (default: {SERIES_RUNNING_DAYS})', default=SERIES_RUNNING_DAYS)
    parser.add_option_group(search_options)

    return parser


def run():

    parser = make_parser()
    command_options, args = parser.parse_args()
    if not args:
        parser.error("no command given, use '-h' for full help")

    command_name, command_args = args[0].lower(), args[1:]

    app_dir = Path.home().joinpath(".videobox")
    os.makedirs(app_dir, exist_ok=True)
    log_dir = app_dir.joinpath("logs")
    os.makedirs(log_dir, exist_ok=True)

    today = date.today()
    log_filename = today.strftime("%Y-%m-%d.log")
    logging.basicConfig(filename=log_dir.joinpath(
        log_filename), level=logging.INFO)
    if videobox.DEBUG or command_options.verbose:
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        for module in ['peewee', 'requests', 'urllib3']:
            # Set higher log level for deps
            logging.getLogger(module).setLevel(logging.INFO)

    config = load_config(app_dir)
    if config.has_option("api", "client_id"):
        client_id = config.get("api", "client_id")
    else:
        # Save client ID for later use
        client_id = uuid.uuid1().hex
        config.add_section("api")
        config.set("api", "client_id", client_id)
        save_config(app_dir, config)

    model.connect(app_dir, should_setup=True)

    try:
        run_command(config, parser, command_name,
                    command_args, command_options)
    except CommandNotFound as ex:
        parser.error(
            f"unrecognized command {command_name}, use '-h' for full help")
    finally:
        model.close()
        logging.shutdown()



from datetime import datetime, date, timedelta


def format_size(value):
    prefix = ['B', 'kB', 'MB', 'GB', 'TB']
    for i in range(len(prefix)):
        if abs(value) < 1000:
            return '%.2f%s' % (value, prefix[i])
        value /= 1000

    return '%.fPB' % value


def format_date(value):
    return value.strftime("%b. %d, %Y")


def format_datetime(value):
    return value.strftime("%b. %d, %Y %H:%M")


def datetime_since(value, comparison_value, default="just now"):
    """
    Returns string representing "time since" e.g.
    3 days ago, 5 hours ago etc.
    """

    diff = comparison_value - value

    periods = (
        (diff.days // 365, "year", "years"),
        (diff.days // 30, "month", "months"),
        (diff.days // 7, "week", "weeks"),
        (diff.days, "day", "days"),
        (diff.seconds // 3600, "hour", "hours"),
        (diff.seconds // 60, "minute", "minutes"),
        (diff.seconds, "second", "seconds"),
    )

    for period, singular, plural in periods:
        if period:
            return "%d %s ago" % (period, singular if period == 1 else plural)

    return default


import requests
import videobox
import logging

TIMEOUT = 10
API_ENDPOINT_URL = "https://videobox.passiomatic.com/1"
USER_AGENT = "Videobox/{0} <https://videobox.passiomatic.com/>".format(
    videobox.__version__)


def get_running_series(client_id):
    return get_url("{0}/series/running?client={1}".format(API_ENDPOINT_URL, client_id))


def get_updated_series(client_id, since):
    return get_url("{0}/series/updated?since={1}&client={2}".format(API_ENDPOINT_URL, int(since.timestamp()), client_id))


def get_series_with_ids(client_id, ids):
    return get_url("{0}/series?ids={1}&client={2}".format(API_ENDPOINT_URL, make_ids(ids), client_id))


def get_series_tags_for_ids(client_id, ids):
    return get_url("{0}/series-tags?ids={1}&client={2}".format(API_ENDPOINT_URL, make_ids(ids), client_id))


def get_episodes_with_ids(client_id, ids):
    return get_url("{0}/episodes?ids={1}&client={2}".format(API_ENDPOINT_URL, make_ids(ids), client_id))


def get_releases_with_ids(client_id, ids):
    return get_url("{0}/releases?ids={1}&client={2}".format(API_ENDPOINT_URL, make_ids(ids), client_id))


def make_ids(ids):
    return ",".join(map(str, ids))


def get_url(url):
    request_headers = {
        'User-Agent': USER_AGENT
    }
    logging.debug(f"Quering API endpoint {url}...")
    return requests.get(url, timeout=TIMEOUT, headers=request_headers)